#!/usr/bin/env python
# Copyright (C) 2014 Elboulangero
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Radio Tray.  If not, see <http://www.gnu.org/licenses/>.

"""
An implementation of the MPRIS D-Bus protocol version 2.
http://specifications.freedesktop.org/mpris-spec/latest/

Greatly inspired by quodlibet MPRIS2 implementation
(almost entirely copy/pasted to say the least)
"""

import dbus
import dbus.service

from dbusutils import DBusIntrospectable, DBusProperty
from dbusutils import dbus_unicode_validate as unival
from events.EventManager import EventManager

# dbus.service.Object must be the last arg given, since Python evaluates
# the class hierarchy right to left, and DBus* override dbus.service.Object
# http://www.ianlewis.org/en/mixins-and-python
class MPRIS2(DBusProperty, DBusIntrospectable, dbus.service.Object):

    BUS_NAME = 'org.mpris.MediaPlayer2.radiotray'
    PATH = '/org/mpris/MediaPlayer2'

    ROOT_IFACE = "org.mpris.MediaPlayer2"

    ROOT_ISPEC = """
<method name="Raise"/>
<method name="Quit"/>"""

    ROOT_PROPS = """
<property name="CanQuit" type="b" access="read"/>
<property name="CanRaise" type="b" access="read"/>
<property name="Fullscreen" type="b" access="readwrite"/>
<property name="CanSetFullscreen" type="b" access="read"/>
<property name="HasTrackList" type="b" access="read"/>
<property name="Identity" type="s" access="read"/>
<property name="DesktopEntry" type="s" access="read"/>
<property name="SupportedUriSchemes" type="as" access="read"/>
<property name="SupportedMimeTypes" type="as" access="read"/>"""

    PLAYER_IFACE = "org.mpris.MediaPlayer2.Player"

    PLAYER_ISPEC = """
<method name="Next"/>
<method name="Previous"/>
<method name="Pause"/>
<method name="PlayPause"/>
<method name="Stop"/>
<method name="Play"/>
<method name="Seek">
  <arg direction="in" name="Offset" type="x"/>
</method>
<method name="SetPosition">
  <arg direction="in" name="TrackId" type="o"/>
  <arg direction="in" name="Position" type="x"/>
</method>
<method name="OpenUri">
  <arg direction="in" name="Uri" type="s"/>
</method>
<signal name="Seeked">
  <arg name="Position" type="x"/>
</signal>"""

    PLAYER_PROPS = """
<property name="PlaybackStatus" type="s" access="read"/>
<property name="LoopStatus" type="s" access="readwrite"/>
<property name="Rate" type="d" access="readwrite"/>
<property name="Shuffle" type="b" access="readwrite"/>
<property name="Metadata" type="a{sv}" access="read"/>
<property name="Volume" type="d" access="readwrite"/>
<property name="Position" type="x" access="read">
  <annotation name="org.freedesktop.DBus.Property.EmitsChangedSignal" 
value="false"/>
</property>
<property name="MinimumRate" type="d" access="read"/>
<property name="MaximumRate" type="d" access="read"/>
<property name="CanGoNext" type="b" access="read"/>
<property name="CanGoPrevious" type="b" access="read"/>
<property name="CanPlay" type="b" access="read"/>
<property name="CanPause" type="b" access="read"/>
<property name="CanSeek" type="b" access="read"/>
<property name="CanControl" type="b" access="read">
  <annotation name="org.freedesktop.DBus.Property.EmitsChangedSignal" \
value="false"/>
</property>"""


    def __init__(self, mediator, systray, event_subscriber):
        DBusIntrospectable.__init__(self)
        DBusProperty.__init__(self)

        self.set_introspection(MPRIS2.ROOT_IFACE, MPRIS2.ROOT_ISPEC)
        self.set_properties(MPRIS2.ROOT_IFACE, MPRIS2.ROOT_PROPS)

        self.set_introspection(MPRIS2.PLAYER_IFACE, MPRIS2.PLAYER_ISPEC)
        self.set_properties(MPRIS2.PLAYER_IFACE, MPRIS2.PLAYER_PROPS)

        bus = dbus.SessionBus()
        name = dbus.service.BusName(MPRIS2.BUS_NAME, bus)
        dbus.service.Object.__init__(self, bus, MPRIS2.PATH, name)

        self.mediator = mediator
        self.systray = systray
        self.event = event_subscriber

        self.event.bind(EventManager.STATE_CHANGED, self.on_state_changed)
        self.event.bind(EventManager.SONG_CHANGED, self.on_song_changed)
        self.event.bind(EventManager.VOLUME_CHANGED, self.on_volume_changed)

    def on_state_changed(self, data):
        self.emit_properties_changed(self.PLAYER_IFACE, ["PlaybackStatus"])

    def on_song_changed(self, data):
        self.emit_properties_changed(self.PLAYER_IFACE, ["Metadata"])

    def on_volume_changed(self, data):
        self.emit_properties_changed(self.PLAYER_IFACE, ["Volume"])

    @dbus.service.method(ROOT_IFACE)
    def Raise(self):
        pass

    @dbus.service.method(ROOT_IFACE)
    def Quit(self):
        self.systray.on_quit(1)

    @dbus.service.method(PLAYER_IFACE)
    def Next(self):
        self.mediator.next()

    @dbus.service.method(PLAYER_IFACE)
    def Previous(self):
        self.mediator.previous()

    @dbus.service.method(PLAYER_IFACE)
    def Pause(self):
        self.mediator.stop()

    @dbus.service.method(PLAYER_IFACE)
    def PlayPause(self):
        if self.mediator.isPlaying():
            self.mediator.stop()
        else:
            self.mediator.playLast()

    @dbus.service.method(PLAYER_IFACE)
    def Stop(self):
        self.mediator.stop()

    @dbus.service.method(PLAYER_IFACE)
    def Play(self):
        if not self.mediator.isPlaying():
            self.mediator.playLast()

    @dbus.service.method(PLAYER_IFACE, in_signature="x")
    def Seek(self, offset):
        pass

    @dbus.service.method(PLAYER_IFACE, in_signature="ox")
    def SetPosition(self, track_id, position):
        pass

    @dbus.service.method(PLAYER_IFACE, in_signature="s")
    def OpenUri(self, uri):
        self.mediator.playUrl(uri)

    @dbus.service.signal(PLAYER_IFACE, signature="x")
    def Seeked(self, position):
        pass

    def get_property(self, interface, name):
        if interface == MPRIS2.ROOT_IFACE:
            if name == "CanQuit":
                return True
            elif name == "CanRaise":
                return False
            elif name == "Fullscreen":
                return False
            elif name == "CanSetFullscreen":
                return False
            elif name == "HasTrackList":
                return False
            elif name == "Identity":
                return "Radio Tray"
            elif name == "DesktopEntry":
                return "radiotray"
            elif name == "SupportedUriSchemes":
                return ["http", "https"]
            elif name == "SupportedMimeTypes":
                return ["audio/*"]

        elif interface == MPRIS2.PLAYER_IFACE:
            if name == "PlaybackStatus":
                if self.mediator.isPlaying():
                    return "Playing"
                else:
                    return "Stopped"
            elif name == "LoopStatus":
                return "None"
            elif name == "Rate":
                return 1.0
            elif name == "Shuffle":
                return False
            elif name == "Metadata":
                metadata = {}
                metadata['mpris:trackid'] = '/org/mpris/MediaPlayer2/CurrentTrack'
                if not self.mediator.isPlaying():
                    return metadata

                context = self.mediator.getContext()
                metadata["xesam:url"] = unival(context.url)
                metadata['xesam:title'] = unival(context.title)
                metadata['xesam:artist'] = unival(context.artist)
                metadata['xesam:album'] = unival(context.album)
                return metadata
            elif name == "Volume":
                return self.mediator.getVolume()
            elif name == "Position":
                return 0
            elif name == "MinimumRate":
                return 1.0
            elif name == "MaximumRate":
                return 1.0
            elif name == "CanGoNext":
                return False
            elif name == "CanGoPrevious":
                return False
            elif name == "CanPlay":
                return True
            elif name == "CanPause":
                return True
            elif name == "CanSeek":
                return False
            elif name == "CanControl":
                return True

            
    def set_property(self, interface, name, value):
        if interface == MPRIS2.ROOT_IFACE:
            if name == "Fullscreen":
                pass

        elif interface == MPRIS2.PLAYER_IFACE:
            if name == "LoopStatus":
                pass
            elif name == "Rate":
                pass
            elif name == "Shuffle":
                pass
            elif name == "Volume":
                if volume < 0 or volume > 100:
                    pass
                self.mediator.set_volume(volume / 100.0)
